package boshclient_test

import (
	"errors"

	"github.com/pivotal-cf/on-demand-service-broker/authorizationheader"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/mockbosh"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("getting deployment", func() {
	var (
		deploymentName   string = "some-deployment"
		manifest         []byte
		deploymentFound  bool
		rawManifest      = []byte("a-raw-manifest")
		manifestFetchErr error
	)

	Context("when the bosh director can be reached", func() {
		JustBeforeEach(func() {
			manifest, deploymentFound, manifestFetchErr = c.GetDeployment(deploymentName, logger)
		})

		Context("when the deployment exists", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Manifest(deploymentName).RespondsWith(rawManifest),
				)
			})

			It("calls the authorization header builder", func() {
				Expect(authHeaderBuilder.BuildCallCount()).To(BeNumerically(">", 0))
			})

			It("returns deployment found", func() {
				Expect(deploymentFound).To(BeTrue())
			})

			It("returns the manifest", func() {
				Expect(manifest).To(Equal(rawManifest))
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})
		})

		Context("when the deployment does not exist", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Manifest(deploymentName).NotFound(),
				)
			})

			It("returns deployment not found", func() {
				Expect(deploymentFound).To(BeFalse())
			})

			It("does not returns the manifest", func() {
				Expect(manifest).To(BeNil())
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})
		})

		Context("when the Authorization header cannot be generated", func() {
			BeforeEach(func() {
				authHeaderBuilder.BuildReturns("", errors.New("some-error"))
			})

			It("returns an error", func() {
				Expect(manifestFetchErr).To(MatchError(ContainSubstring("some-error")))
			})
		})
	})

	Context("when the BOSH director cannot be reached", func() {
		It("returns a bosh request error", func() {
			c, err := boshclient.New("http://localhost", authorizationheader.NewBasicAuthHeaderBuilder("", ""), false, nil)
			Expect(err).NotTo(HaveOccurred())

			_, deploymentFound, manifestFetchErr = c.GetDeployment(deploymentName, logger)

			Expect(manifestFetchErr).To(MatchError(ContainSubstring("error reaching bosh director")))
			Expect(manifestFetchErr).To(BeAssignableToTypeOf(boshclient.RequestError{}))
			Expect(deploymentFound).To(BeFalse())
		})
	})
})
