package broker

import (
	"context"
	"encoding/json"
	"fmt"
	"log"

	"github.com/pborman/uuid"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/brokercontext"
)

func (b *Broker) Provision(
	ctx context.Context,
	instanceID string,
	details brokerapi.ProvisionDetails,
	asyncAllowed bool,
) (brokerapi.ProvisionedServiceSpec, error) {
	b.deploymentLock.Lock()
	defer b.deploymentLock.Unlock()

	requestID := uuid.New()
	ctx = brokercontext.New(ctx, string(OperationTypeCreate), requestID, b.serviceOffering.Name, instanceID)
	logger := b.loggerFactory.NewLoggerWithContext(ctx)

	if !asyncAllowed {
		return brokerapi.ProvisionedServiceSpec{}, brokerapi.ErrAsyncRequired
	}

	detailsWithRawParameters := brokerapi.DetailsWithRawParameters(details)
	requestParams, err := convertDetailsToMap(detailsWithRawParameters)
	if err != nil {
		return brokerapi.ProvisionedServiceSpec{}, err
	}

	boshTaskID, boshContextID, dashboardURL, displayableErr := b.provisionInstance(
		ctx,
		instanceID,
		details,
		requestParams,
		logger,
	)
	if displayableErr.Occurred() {
		logger.Println(displayableErr)
		return brokerapi.ProvisionedServiceSpec{}, displayableErr.ErrorForCFUser()
	}

	operationData, err := json.Marshal(OperationData{
		BoshTaskID:    boshTaskID,
		OperationType: OperationTypeCreate,
		BoshContextID: boshContextID,
		PlanID:        details.PlanID,
	})
	if err != nil {
		return brokerapi.ProvisionedServiceSpec{}, err
	}

	return brokerapi.ProvisionedServiceSpec{
		IsAsync:       true,
		DashboardURL:  dashboardURL,
		OperationData: string(operationData),
	}, nil
}

func (b *Broker) provisionInstance(
	ctx context.Context,
	instanceID string,
	details brokerapi.ProvisionDetails,
	requestParams map[string]interface{},
	logger *log.Logger,
) (int, string, string, DisplayableError) {
	errs := func(err DisplayableError) (int, string, string, DisplayableError) {
		return 0, "", "", err
	}

	plan, found := b.serviceOffering.FindPlanByID(details.PlanID)
	if !found {
		return errs(NewDisplayableError(
			fmt.Errorf("plan %s not found", details.PlanID),
			fmt.Errorf("finding plan ID %s", details.PlanID),
		))
	}

	_, found, err := b.boshClient.GetDeployment(deploymentName(instanceID), logger)
	switch err := err.(type) {
	case boshclient.RequestError:
		return errs(NewBoshRequestError("create", fmt.Errorf("could not get manifest: %s", err)))
	case error:
		return errs(NewGenericError(ctx, fmt.Errorf("could not get manifest: %s", err)))
	}

	if found {
		return errs(NewDisplayableError(
			brokerapi.ErrInstanceAlreadyExists,
			fmt.Errorf("deploying instance %s", instanceID),
		))
	}

	var planCounts map[string]int
	// check if the global service instance limit has been reached
	if b.serviceOffering.GlobalQuotas.ServiceInstanceLimit != nil {
		var displayableError DisplayableError
		planCounts, displayableError = b.checkGlobalQuota(ctx, b.serviceOffering.ID, logger)
		if displayableError.Occurred() {
			return errs(displayableError)
		}
	}

	// check if the quota for the plan has been reached
	if plan.Quotas.ServiceInstanceLimit != nil {
		limit := *plan.Quotas.ServiceInstanceLimit
		planCount, displayableError := b.getPlanCount(ctx, details.PlanID, planCounts, logger)
		if displayableError.Occurred() {
			return errs(displayableError)
		}

		if planCount >= limit {
			return errs(NewDisplayableError(
				brokerapi.ErrPlanQuotaExceeded,
				fmt.Errorf("plan quota exceeded for plan ID %s", details.PlanID),
			))
		}
	}

	// generateBoshContextID
	var boshContextID string
	if plan.LifecycleErrands != nil {
		boshContextID = uuid.New()
	}

	// do the deploy!
	boshTaskID, manifest, err := b.deployer.Deploy(instanceID, plan.ID, deploymentName(instanceID), OperationTypeCreate, requestParams, nil, nil, boshContextID, logger)
	switch err := err.(type) {
	case boshclient.RequestError:
		return errs(NewBoshRequestError("create", err))
	case DisplayableError:
		return errs(err)
	case error:
		return errs(NewGenericError(ctx, err))
	}

	ctx = brokercontext.WithBoshTaskID(ctx, boshTaskID)

	abridgedPlan := plan.AdapterPlan(b.serviceOffering.GlobalProperties)

	dashboardUrl, err := b.serviceAdapter.GenerateDashboardUrl(instanceID, abridgedPlan, manifest, logger)
	switch err := err.(type) {
	case ServiceAdapterCommandNotImplementedError:
		return boshTaskID, boshContextID, dashboardUrl, DisplayableError{}
	case DisplayableError:
		return errs(err)
	case error:
		return errs(NewGenericError(ctx, err))
	}

	return boshTaskID, boshContextID, dashboardUrl, DisplayableError{}
}

func (b *Broker) getPlanCount(ctx context.Context, planID string, planCounts map[string]int, logger *log.Logger) (int, DisplayableError) {
	var planCount int

	if planCounts != nil {
		planCount = planCounts[planID]
	} else {
		var countErr error
		planCount, countErr = b.cfClient.CountInstancesOfPlan(b.serviceOffering.ID, planID, logger)
		if countErr != nil {
			return 0, NewGenericError(ctx, fmt.Errorf("could not count instances of plan: %s", countErr))
		}
	}

	return planCount, DisplayableError{}
}

// b.checkGlobalQuota counts the total number of service instances of a service offering and returns an error if the
// quota has been reached, or if the instances could not be counted.
func (b *Broker) checkGlobalQuota(
	ctx context.Context,
	serviceOfferingID string,
	logger *log.Logger,
) (map[string]int, DisplayableError) {

	planCounts, err := b.cfClient.CountInstancesOfPlans(serviceOfferingID, logger)
	if err != nil {
		return nil, NewGenericError(ctx, err)
	}

	var totalServiceInstances = 0
	for _, count := range planCounts {
		totalServiceInstances += count
	}

	if b.serviceOffering.GlobalQuotas.ServiceInstanceLimit != nil && totalServiceInstances >= *b.serviceOffering.GlobalQuotas.ServiceInstanceLimit {
		return nil, NewDisplayableError(
			brokerapi.ErrServiceQuotaExceeded,
			fmt.Errorf("service quota exceeded for service ID %s", b.serviceOffering.ID),
		)
	}

	return nil, DisplayableError{}
}
