package broker

import "github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"

type ServiceOffering struct {
	ID               string
	Name             string `yaml:"service_name"`
	Description      string `yaml:"service_description"`
	Bindable         bool
	PlanUpdatable    bool     `yaml:"plan_updatable"`
	Requires         []string `yaml:"requires,omitempty"`
	Metadata         ServiceMetadata
	DashboardClient  *DashboardClient `yaml:"dashboard_client,omitempty"`
	Tags             []string
	GlobalProperties serviceadapter.Properties `yaml:"global_properties"`
	GlobalQuotas     Quotas                    `yaml:"global_quotas"`
	Plans            Plans
}

func (s ServiceOffering) FindPlanByID(id string) (Plan, bool) {
	return s.Plans.FindByID(id)
}

func (s ServiceOffering) HasLifecycleErrands() bool {
	for _, plan := range s.Plans {
		if plan.LifecycleErrands != nil {
			if plan.LifecycleErrands.PostDeploy != "" || plan.LifecycleErrands.PreDelete != "" {
				return true
			}
		}
	}

	return false
}

type Plans []Plan

func (p Plans) FindByID(id string) (Plan, bool) {
	for _, plan := range p {
		if plan.ID == id {
			return plan, true
		}
	}
	return Plan{}, false
}

type Plan struct {
	ID               string `yaml:"plan_id"`
	Name             string
	Free             *bool
	Bindable         *bool
	Description      string
	Metadata         PlanMetadata
	Quotas           Quotas `yaml:"quotas,omitempty"`
	Properties       serviceadapter.Properties
	InstanceGroups   []serviceadapter.InstanceGroup `yaml:"instance_groups,omitempty"`
	Update           *serviceadapter.Update         `yaml:"update,omitempty"`
	LifecycleErrands *LifecycleErrands              `yaml:"lifecycle_errands,omitempty"`
}

func (p Plan) AdapterPlan(globalProperties serviceadapter.Properties) serviceadapter.Plan {
	return serviceadapter.Plan{
		Properties:     mergeProperties(p.Properties, globalProperties),
		InstanceGroups: p.InstanceGroups,
		Update:         p.Update,
	}
}

func mergeProperties(planProperties, globalProperties serviceadapter.Properties) serviceadapter.Properties {
	properties := serviceadapter.Properties{}
	// avoid modifying global properties
	for k, v := range globalProperties {
		properties[k] = v
	}
	for k, v := range planProperties {
		properties[k] = v
	}
	return properties
}

func (p Plan) PostDeployErrand() string {
	if p.LifecycleErrands == nil {
		return ""
	}

	return p.LifecycleErrands.PostDeploy
}

func (p Plan) PreDeleteErrand() string {
	if p.LifecycleErrands == nil {
		return ""
	}

	return p.LifecycleErrands.PreDelete
}

type LifecycleErrands struct {
	PostDeploy string `yaml:"post_deploy"`
	PreDelete  string `yaml:"pre_delete"`
}

type PlanMetadata struct {
	DisplayName string     `yaml:"display_name"`
	Bullets     []string   `yaml:"bullets,omitempty"`
	Costs       []PlanCost `yaml:"costs"`
}

type PlanCost struct {
	Amount map[string]float64 `yaml:"amount"`
	Unit   string             `yaml:"unit"`
}

type ServiceMetadata struct {
	DisplayName         string `yaml:"display_name"`
	ImageURL            string `yaml:"image_url"`
	LongDescription     string `yaml:"long_description"`
	ProviderDisplayName string `yaml:"provider_display_name"`
	DocumentationURL    string `yaml:"documentation_url"`
	SupportURL          string `yaml:"support_url"`
}

type DashboardClient struct {
	ID          string
	Secret      string
	RedirectUri string `yaml:"redirect_uri"`
}

type Quotas struct {
	ServiceInstanceLimit *int `yaml:"service_instance_limit,omitempty"`
}
