package broker

import (
	"context"
	"encoding/json"
	"fmt"

	"errors"

	"github.com/pborman/uuid"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/brokercontext"
)

func (b *Broker) Update(
	ctx context.Context,
	instanceID string,
	details brokerapi.UpdateDetails,
	asyncAllowed bool,
) (brokerapi.UpdateServiceSpec, error) {
	b.deploymentLock.Lock()
	defer b.deploymentLock.Unlock()

	requestID := uuid.New()
	ctx = brokercontext.New(ctx, string(OperationTypeUpdate), requestID, b.serviceOffering.Name, instanceID)
	logger := b.loggerFactory.NewLoggerWithContext(ctx)

	if !asyncAllowed {
		return brokerapi.UpdateServiceSpec{}, brokerapi.ErrAsyncRequired
	}

	errs := func(err DisplayableError) (brokerapi.UpdateServiceSpec, error) {
		logger.Println(err)
		return brokerapi.UpdateServiceSpec{IsAsync: true}, err.ErrorForCFUser()
	}

	plan, found := b.serviceOffering.FindPlanByID(details.PlanID)
	if !found {
		return errs(NewDisplayableError(fmt.Errorf("plan %s not found", details.PlanID), fmt.Errorf("finding plan ID %s", details.PlanID)))
	}

	if details.PreviousValues.PlanID != plan.ID {
		if err := b.validatePlanQuota(ctx, details.ServiceID, plan, logger); err != NilError {
			return errs(err)
		}
	}

	logger.Printf("updating instance %s", instanceID)
	detailsWithRawParameters := brokerapi.DetailsWithRawParameters(details)
	detailsMap, err := convertDetailsToMap(detailsWithRawParameters)
	if err != nil {
		return errs(NewGenericError(ctx, err))
	}

	var boshContextID string
	var operationDataPlanID string
	if plan.LifecycleErrands != nil {
		boshContextID = uuid.New()
		operationDataPlanID = details.PlanID
	}

	deployerFunc, err := b.chooseDeployerFunc(detailsMap)
	if err != nil {
		return errs(NewGenericError(ctx, err))
	}

	boshTaskID, err := b.redeploy(
		instanceID,
		details.PlanID,
		OperationTypeUpdate,
		detailsMap,
		&details.PreviousValues.PlanID,
		boshContextID,
		deployerFunc,
		logger,
	)

	switch err := err.(type) {
	case boshclient.RequestError:
		return errs(NewBoshRequestError("update", fmt.Errorf("error deploying instance: %s", err)))
	case DisplayableError:
		return errs(err)
	case error:
		return errs(
			NewGenericError(ctx, fmt.Errorf("error deploying instance: %s", err)),
		)
	}

	ctx = brokercontext.WithBoshTaskID(ctx, boshTaskID)

	operationData, err := json.Marshal(OperationData{
		BoshTaskID:    boshTaskID,
		OperationType: OperationTypeUpdate,
		BoshContextID: boshContextID,
		PlanID:        operationDataPlanID,
	})
	if err != nil {
		return errs(NewGenericError(ctx, err))
	}

	return brokerapi.UpdateServiceSpec{IsAsync: true, OperationData: string(operationData)}, nil
}

func (b *Broker) chooseDeployerFunc(detailsMap map[string]interface{}) (DeployerFunc, error) {
	var applyChanges bool

	parameters, _ := detailsMap["parameters"].(map[string]interface{})
	applyChangesRaw := parameters["apply-changes"]
	if applyChangesRaw != nil {
		var ok bool
		applyChanges, ok = parameters["apply-changes"].(bool)
		if !ok {
			return nil, errors.New("update called with apply-changes set to non-boolean")
		}
	}
	if applyChanges {
		return b.deployer.Deploy, nil
	}

	return b.deployer.Update, nil
}
