package task

import (
	"errors"
	"fmt"
	"log"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

//go:generate counterfeiter -o fakes/fake_bosh_client.go . BoshClient
type BoshClient interface {
	Deploy(manifest []byte, contextID string, logger *log.Logger) (int, error)
	GetTasks(deploymentName string, logger *log.Logger) (boshclient.BoshTasks, error)
}

//go:generate counterfeiter -o fakes/fake_manifest_generator.go . ManifestGenerator
type ManifestGenerator interface {
	GenerateManifest(deploymentName, instanceID, planID string, requestParams map[string]interface{}, oldManifest []byte, previousPlanID *string, logger *log.Logger) (BoshManifest, error)
}

//go:generate counterfeiter -o fakes/fake_feature_flags.go . FeatureFlags
type FeatureFlags interface {
	CFUserTriggeredUpgrades() bool
}

type deployer struct {
	boshClient        BoshClient
	manifestGenerator ManifestGenerator
	featureFlags      FeatureFlags
}

func NewDeployer(
	boshClient BoshClient,
	manifestGenerator ManifestGenerator,
	featureFlags FeatureFlags,
) deployer {
	return deployer{
		boshClient:        boshClient,
		manifestGenerator: manifestGenerator,
		featureFlags:      featureFlags,
	}
}

func (d deployer) Deploy(
	instanceID, planID, deploymentName string,
	operationType broker.OperationType,
	requestParams map[string]interface{},
	oldManifest []byte,
	previousPlanID *string,
	boshContextID string,
	logger *log.Logger,
) (int, []byte, error) {

	err := d.assertNoOperationsInProgress(deploymentName, logger)
	if err != nil {
		return 0, nil, err
	}

	err = d.assertValidRequest(requestParams, planID, previousPlanID)
	if err != nil {
		return 0, nil, err
	}

	manifest, err := d.manifestGenerator.GenerateManifest(
		deploymentName,
		instanceID,
		planID,
		requestParams,
		oldManifest,
		previousPlanID,
		logger,
	)
	if err != nil {
		return 0, nil, err
	}

	taskID, err := d.boshClient.Deploy(manifest, boshContextID, logger)
	if err != nil {
		return 0, nil, fmt.Errorf("error deploying instance: %s\n", err)
	}
	logger.Printf("Bosh task ID for %s instance %s is %d\n", operationType, instanceID, taskID)

	return taskID, manifest, nil
}

func (d deployer) Update(
	instanceID, planID, deploymentName string,
	operationType broker.OperationType,
	requestParams map[string]interface{},
	oldManifest []byte,
	previousPlanID *string,
	boshContextID string,
	logger *log.Logger,
) (int, []byte, error) {

	err := d.assertNoOperationsInProgress(deploymentName, logger)
	if err != nil {
		return 0, nil, err
	}

	err = d.assertValidRequest(requestParams, planID, previousPlanID)
	if err != nil {
		return 0, nil, err
	}

	pendingChanges, pendingChangesErr := d.hasPendingChanges(deploymentName, instanceID, previousPlanID, operationType, oldManifest, logger)
	if pendingChangesErr != nil {
		return 0, nil, pendingChangesErr
	}

	if pendingChanges == true {
		err = errors.New("pending changes detected")
		if d.featureFlags.CFUserTriggeredUpgrades() {
			return 0, nil, broker.NewPendingChangesError(err)
		} else {
			return 0, nil, broker.NewApplyChangesDisabledError(err)
		}
	}

	manifest, err := d.manifestGenerator.GenerateManifest(deploymentName, instanceID, planID, requestParams, oldManifest, previousPlanID, logger)
	if err != nil {
		return 0, nil, err
	}

	taskID, err := d.boshClient.Deploy(manifest, boshContextID, logger)
	if err != nil {
		return 0, nil, fmt.Errorf("error deploying instance: %s\n", err)
	}
	logger.Printf("Bosh task ID for %s instance %s is %d\n", operationType, instanceID, taskID)

	return taskID, manifest, nil
}

func (d deployer) assertNoOperationsInProgress(deploymentName string, logger *log.Logger) error {
	tasks, err := d.boshClient.GetTasks(deploymentName, logger)
	if err != nil {
		return fmt.Errorf("error getting tasks for deployment %s: %s\n", deploymentName, err)
	}

	if incompleteTasks := tasks.IncompleteTasks(); len(incompleteTasks) != 0 {
		userError := errors.New("An operation is in progress for your service instance. Please try again later.")
		operatorError := broker.NewOperationInProgressError(
			fmt.Errorf("deployment %s is still in progress: tasks %s\n",
				deploymentName,
				incompleteTasks.ToLog()),
		)
		return broker.NewDisplayableError(userError, operatorError)
	}

	return nil
}

func (d deployer) assertValidRequest(requestParams map[string]interface{}, planID string, previousPlanID *string) error {
	parameters, ok := requestParams["parameters"].(map[string]interface{})
	if !ok {
		return nil
	}

	return d.assertValidApplyChangesRequest(parameters, planID, previousPlanID)
}

func (d deployer) assertValidApplyChangesRequest(parameters map[string]interface{}, planID string, previousPlanID *string) error {
	if parameters["apply-changes"] == nil {
		return nil
	}

	applyChanges, ok := parameters["apply-changes"].(bool)
	if !ok {
		err := errors.New("update called with apply-changes set to non-boolean")
		if d.featureFlags.CFUserTriggeredUpgrades() {
			return broker.NewPendingChangesError(err)
		} else {
			return broker.NewApplyChangesNotPermittedError(err)
		}
	}

	if !applyChanges {
		return nil
	}

	if !d.featureFlags.CFUserTriggeredUpgrades() {
		return broker.NewApplyChangesNotPermittedError(errors.New("'cf_user_triggered_upgrades' feature is disabled"))
	}

	if previousPlanID != nil && planID != *previousPlanID {
		err := errors.New("update called with apply-changes and a plan change")
		return broker.NewPendingChangesError(err)
	}

	if len(parameters) > 1 {
		err := errors.New("update called with apply-changes and arbitrary parameters set")
		return broker.NewPendingChangesError(err)
	}

	delete(parameters, "apply-changes")

	return nil
}

func (d deployer) hasPendingChanges(deploymentName, instanceID string, previousPlanID *string, operationType broker.OperationType, oldManifest BoshManifest, logger *log.Logger) (bool, error) {
	// generate a manifest without params and the previously used plan ID
	regeneratedManifest, err := d.manifestGenerator.GenerateManifest(deploymentName, instanceID, *previousPlanID, map[string]interface{}{}, oldManifest, previousPlanID, logger)
	if err != nil {
		return false, err
	}

	// check whether the manifest generated by the adapter is the same we previously deployed
	manifestsEqual, err := regeneratedManifest.Equals(oldManifest)
	if err != nil {
		return false, fmt.Errorf("error detecting change in manifest: %s", err)
	}

	return !manifestsEqual, nil
}
