package upgrader

import (
	"encoding/json"
	"fmt"
	"net/http"
	"net/url"
	"time"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"
)

type upgrader struct {
	brokerUsername  string
	brokerPassword  string
	brokerUrl       string
	pollingInterval time.Duration
}

type result string

const (
	resultSuccess  result = "success"
	resultRetry    result = "retry"
	resultNotFound result = "not found"
	resultOrphan   result = "orphan"
	resultError    result = "error"
)

func New(brokerUsername, brokerPassword, brokerUrl string, pollingIntervalMilliseconds int64) upgrader {
	return upgrader{
		brokerUsername:  brokerUsername,
		brokerPassword:  brokerPassword,
		brokerUrl:       brokerUrl,
		pollingInterval: time.Duration(pollingIntervalMilliseconds) * time.Millisecond,
	}
}

func (u upgrader) Upgrade() ([]string, []string, error) {
	var upgraded, orphans []string

	instancesToUpgrade, err := u.getServiceInstances()
	if err != nil {
		return nil, nil, fmt.Errorf("error getting service instances: %s\n", err)
	}

	for len(instancesToUpgrade) != 0 {
		upgradedInstances, orphanInstances, retryInstances, err := u.upgradeInstances(instancesToUpgrade)
		if err != nil {
			return nil, nil, err
		}

		upgraded = append(upgraded, upgradedInstances...)
		orphans = append(orphans, orphanInstances...)
		instancesToUpgrade = retryInstances

		if len(instancesToUpgrade) > 0 {
			time.Sleep(u.pollingInterval)
		}
	}

	return upgraded, orphans, nil
}

func (u upgrader) upgradeInstances(instances []string) ([]string, []string, []string, error) {
	var upgraded, orphan, retry []string

	for _, instance := range instances {
		operationData, result, err := u.upgradeInstance(instance)
		if err != nil {
			return nil, nil, nil, fmt.Errorf("Upgrade failed for service_instance_guid %s: %s\n", instance, err)
		}

		switch result {
		case resultOrphan:
			orphan = append(orphan, instance)
		case resultRetry:
			retry = append(retry, instance)
		case resultSuccess:
			if err := u.pollLastOperation(instance, operationData); err != nil {
				return nil, nil, nil, err
			}
			upgraded = append(upgraded, instance)
		}
	}

	return upgraded, orphan, retry, nil
}

func (u upgrader) pollLastOperation(instance string, opData broker.OperationData) error {
	for {
		time.Sleep(u.pollingInterval)

		lastOperation, err := u.getLastOperation(instance, opData)
		if err != nil {
			return fmt.Errorf("error getting last operation: %s\n", err)
		}

		switch lastOperation.State {
		case brokerapi.Failed:
			return fmt.Errorf("Upgrade failed for service_instance_guid %s: %s", instance, lastOperation.Description)
		case brokerapi.Succeeded:
			return nil
		}
	}
}

func (u upgrader) getServiceInstances() ([]string, error) {
	var instances []mgmtapi.Instance
	if err := u.makeRequest("/mgmt/service_instances", &instances); err != nil {
		return nil, err
	}

	var instanceIDs []string
	for _, instance := range instances {
		instanceIDs = append(instanceIDs, instance.InstanceID)
	}
	return instanceIDs, nil
}

func (u upgrader) upgradeInstance(instance string) (broker.OperationData, result, error) {
	request, err := http.NewRequest("PATCH", fmt.Sprintf("%s%s", u.brokerUrl, "/mgmt/service_instances/"+instance), nil)
	if err != nil {
		return broker.OperationData{}, resultError, err
	}

	request.SetBasicAuth(u.brokerUsername, u.brokerPassword)

	response, err := http.DefaultClient.Do(request)
	if err != nil {
		return broker.OperationData{}, resultError, err
	}
	defer response.Body.Close()

	switch response.StatusCode {
	case http.StatusAccepted:
		var operationData broker.OperationData
		if err = json.NewDecoder(response.Body).Decode(&operationData); err != nil {
			return broker.OperationData{}, resultError, fmt.Errorf("can't parse response %s", err)
		}

		return operationData, resultSuccess, nil
	case http.StatusNotFound:
		return broker.OperationData{}, resultNotFound, nil
	case http.StatusGone:
		return broker.OperationData{}, resultOrphan, nil
	case http.StatusConflict:
		return broker.OperationData{}, resultRetry, nil
	}

	var errorResponse brokerapi.ErrorResponse
	if err := json.NewDecoder(response.Body).Decode(&errorResponse); err != nil {
		return broker.OperationData{}, resultError, fmt.Errorf("can't parse response %s", err)
	}

	return broker.OperationData{}, resultError, fmt.Errorf(
		"unexpected status code: %d. description: %s", response.StatusCode, errorResponse.Description,
	)
}

func (u upgrader) getLastOperation(instance string, opData broker.OperationData) (brokerapi.LastOperation, error) {
	opDataJSON, err := json.Marshal(opData)
	if err != nil {
		return brokerapi.LastOperation{}, err
	}

	operationQueryStr := url.QueryEscape(string(opDataJSON))

	var lastOperation brokerapi.LastOperation
	if err := u.makeRequest("/v2/service_instances/"+instance+"/last_operation?operation="+operationQueryStr, &lastOperation); err != nil {
		return brokerapi.LastOperation{}, err
	}

	return lastOperation, nil
}

func (u upgrader) makeRequest(requestPath string, responseData interface{}) error {
	request, err := http.NewRequest("GET", fmt.Sprintf("%s%s", u.brokerUrl, requestPath), nil)
	if err != nil {
		return err
	}

	request.SetBasicAuth(u.brokerUsername, u.brokerPassword)

	response, err := http.DefaultClient.Do(request)
	if err != nil {
		return err
	}

	if response.StatusCode != 200 {
		return fmt.Errorf("response returned bad status: %s", response.Status)
	}

	err = json.NewDecoder(response.Body).Decode(responseData)
	if err != nil {
		return err
	}

	return nil
}
